#include "cset.hh"
#include "encoders.hh"
#include "config.hh"

#include <cwchar>
#include <map>

void Chain::AddEncoding(const std::string& s)
{
    try
    {
        const Encoding& enc = FindEncoder(s);
        chain.push_back(&enc);
    }
    catch(const std::string& )
    {
        std::fprintf(stderr, "Encoding '%s' is unknown\n", s.c_str());
    }
}

const std::string ChainOut::Encode(const std::wstring& s) const
{
    std::string result;

    int shift = 0;
    unsigned old_enc = 0;

    for(unsigned a=0; a<s.size(); ++a)
    {
        /* Find the first encoder able to generate the result */
        for(unsigned e=0; e<chain.size(); ++e)
        {
            bool last = (e+1) == chain.size();

            /* Calculate the shift from this chain's opinion */
            const int shift_of_this_chain = e==old_enc ? shift : 0;

            if(!chain[e]->Encodable(s[a], shift_of_this_chain))
            {
                // Wasn't encodable. If we're not desperate, try the next option.
                if(!last)
                {
                    continue;
                }
            }

            // Either it worked or we're desperate.

            /* If we had a shift from a different encoder, unshift it */
            if(shift && e != old_enc)
            {
                chain[old_enc]->SeqEnd(result);
                shift = 0;
            }

            /* Encode it. This call may change shift. */
            chain[e]->SeqValue(result, s[a], shift);

            /* Remember the encoder */
            if(shift)
            {
                old_enc = e;
            }

            break;
        }
        /* FIXME: We don't know if any of the encoders actually did it. */
    }
    if(shift)
    {
        /* Unshift the last shift */
        chain[old_enc]->SeqEnd(result);
        shift = 0;
    }
    return result;
}

const std::wstring ChainIn::Decode(const std::string& s) const
{
    std::wstring result;

    Encoding::SeqResult shifted;

    for(unsigned a=0; a<s.size(); )
    {
        try
        {
            /* If we had a shift going on, use the
             * same encoder as before.
             */
            if(shifted)
            {
                unsigned e = shifted.GetValue();
                const Encoding& enc = *chain[e];

                /* If the old encoder found a new shift? */
                Encoding::SeqResult tmp = enc.IsShift(s, a);
                if(tmp)
                {
                    /* Yes - eat and process it. */
                    a += tmp.GetLength();
                    tmp.SetValue(e);
                    throw tmp;
                }
                /* No changes in shiftiness. Then try decoding. */
                tmp = enc.IsData(s, a, shifted.GetShiftType());
                if(tmp)
                {
                    /* A character found. */
                    a      += tmp.GetLength();
                    result += (wchar_t)tmp.GetValue();
                    continue;
                }
                /* This encoder couldn't decode the character! */
                /* This is unexpected. So we'll reset the shift status
                 * and try all decoders normally.
                 */
                shifted.SetBool(false);
            }

            /* The last choice is used if nothing else works */
            unsigned BestCandidate = chain.size()-1;
            int BestFondness       = -1;

            /* Otherwise find one that is best suited */
            for(unsigned e=0; e<chain.size(); ++e)
            {
                /* If there is a shift sequence, it overrides the priority. */
                Encoding::SeqResult tmp = chain[e]->IsShift(s, a);
                if(tmp)
                {
                    /* Yes - eat and process it. */
                    a += tmp.GetLength();
                    tmp.SetValue(e);
                    throw tmp;
                }

                /* 0 = no shift. */
                tmp = chain[e]->IsData(s, a, 0);
                if(tmp)
                {
                    /* This encoder was able to decode it. */
                    int fondness = chain[e]->GetFondness();
                    if(fondness > BestFondness)
                    {
                        BestFondness  = fondness;
                        BestCandidate = e;
                    }
                }
            }

            /* So the "best choice" will now have to decode it. */
            Encoding::SeqResult best = chain[BestCandidate]->IsData(s, a, 0);

            a      += best.GetLength();
            result += (wchar_t)best.GetValue();
        }
        catch(const Encoding::SeqResult foundshift)
        {
            int shift = foundshift.GetShiftType();
            if(shift == 0)
                shifted.SetBool(false);
            else
                shifted = foundshift;
            continue;
        }
    }
    /* Is there still an shift? */
    /* If there is, we don't care. A new line is a completely new line. */
    return result;
}

static class ChainSetup
{
private:
    std::map<std::string, Chain2> chains;
private:
    void DefineChain(const std::string& name,
                     const ChainIn& chain_in,
                     const ChainOut& chain_out)
    {
        Chain2 tmp(chain_in, chain_out, name);
        chains[name] = tmp;
    }
public:
    ChainSetup(): chains()
    {
        Reload();
    }
    void Reload()
    {
        chains.clear();

        ConfParser::FieldType settings;

        std::map<std::string, ChainIn> input_chains;
        std::map<std::string, ChainOut> output_chains;

        settings = GetConf("input_chains", "define_in_chain");
        for(unsigned a=0; a<settings.size(); ++a)
        {
            const ConfParser::FuncType& func = settings[a];

            if(func.size() <= 1)
            {
                continue;
            }

            ChainIn chain;
            chain.name = func[0].SField;
            for(unsigned b=1; b<func.size(); ++b)
            {
                chain.AddEncoding(func[b].SField);
            }
            input_chains[func[0].SField] = chain;
        }

        settings = GetConf("output_chains", "define_out_chain");
        for(unsigned a=0; a<settings.size(); ++a)
        {
            const ConfParser::FuncType& func = settings[a];

            if(func.size() <= 1)
            {
                continue;
            }

            ChainOut chain;
            chain.name = func[0].SField;
            for(unsigned b=1; b<func.size(); ++b)
            {
                chain.AddEncoding(func[b].SField);
            }
            output_chains[func[0].SField] = chain;
        }

        settings = GetConf("chains", "define_chain");
        for(unsigned a=0; a<settings.size(); ++a)
        {
            const ConfParser::FuncType& func = settings[a];

            if(func.size() != 3)
            {
                continue;
            }

            DefineChain(func[0].SField, input_chains[func[1].SField],
                                        output_chains[func[2].SField]
                       );
        }
    }

    const Chain2& FindChain(const std::string& chainname) const
    {
        std::map<std::string, Chain2>::const_iterator i;
        i = chains.find(chainname);
        if(i != chains.end()) return i->second;

        throw chainname;
    }
} ChainSetup;

void ReloadChainSetup()
{
    ChainSetup.Reload();
}

const Chain2& FindChain2(const std::string& chainname)
{
    return ChainSetup.FindChain(chainname);
}

const Chain2 CreateBuiltinChainPair(const std::string& encname)
{
    Chain2 result;
    result.in.AddEncoding(encname);
    result.out.AddEncoding(encname);
    return result;
}

const std::string Chain::Explain() const
{
    std::string result = name + "[";
    for(unsigned a=0; a<chain.size(); ++a)
    {
        if(a>0)result += ' ';
        result += chain[a]->name;
    }
    result += "]";
    return result;
}
